import pymana.utils as utils
# from skimage.color import rgb2gray
from skimage.color import rgb2hed
from skimage.exposure import rescale_intensity
from skimage.util import invert
# from skimage.filters import gaussian, median
import dask
import numpy as np
from sympy import diff, solveset, Interval
from sympy.abc import s
import cv2
# from vispy.color import Colormap


def get_hematoxylin(rgb):
    """Run color deconvolution on an Aperio SVS image.

        Parameters
        ----------
        rgb
                A dask array representing the RGB image to be processed.
    """
    # matplotlib navy is (22, 0, 134), vispy navy is (0, 0, 128)
    # cmap_hema = Colormap(['white', 'navy'])

    # matplotlib saddlebrown is (144, 66, 0), vispy saddlebrown is (139, 69, 19)
    # cmap_dab = Colormap(['white', 'saddlebrown'])

    # matplotlib darkviolet is (166, 0, 218), vispy darkviolet is (148, 0, 211)
    # cmap_eosin = Colormap(['darkviolet', 'white'])

    ihc_hed = rgb2hed(rgb)
    arr_hema = ihc_hed[:, :, 0]
    return arr_hema


def candidate_thresholds(img):
    # convert image to grayscale (eliminate hue + saturation, retain luminance)
    @dask.delayed(pure=True)
    def get_grayscale(rgb):
        # grayscale = rgb2gray(rgb)
        grayscale = get_hematoxylin(rgb)
        grayscale = invert(rescale_intensity(grayscale, out_range=(0, 255)))
        # if a filter is used, grayscale should be:
        # rescale_intensity(invert(grayscale), out_range=(0, 255))
        # grayscale = median(grayscale)
        return grayscale
    grayscale = get_grayscale(img)

    # compute histogram
    N = 255
    h, bins = dask.array.histogram(grayscale, bins=np.arange(N + 1))
    histogram = dask.delayed(dict, pure=True)(
        dask.delayed(zip, pure=True)(bins, h))

    # compute progressive weighted mean (PWM_CURVE)
    @dask.delayed(pure=True)
    def calculate_pwm_curve(histogram):
        sum_wx, sum_w = [0], [histogram[0]]
        for P in range(1, N):
            sum_wx.append(sum_wx[P - 1] + histogram[P] * P)
            sum_w.append(sum_w[P - 1] + histogram[P])
        with np.errstate(divide='ignore', invalid='ignore'):
            pwm_curve = dict(
                zip(bins, np.nan_to_num(np.divide(sum_wx, sum_w))))
        return pwm_curve
    pwm_curve = calculate_pwm_curve(histogram)

    @dask.delayed(pure=True)
    def get_inflection_points(pwm_curve):
        # fit 15th order polynomial to PWM_CURVE
        coef = np.polynomial.polynomial.polyfit(
            list(pwm_curve.keys()), list(pwm_curve.values()), 15)
        # instead of the line under this, take a look at full param in polyfit
        function = sum(co * s**i for i, co in enumerate(coef))
        # some of these might not be inflection points:
        inflection_points = solveset(
            diff(diff(function)), domain=Interval(0, N))
        return list(inflection_points)
    inflection_points = get_inflection_points(pwm_curve)

    # return (grayscale image as Dask array, inflection points as list)
    return grayscale, inflection_points


def median_areas(image):
    # get candidate thresholds
    dask_gs, dask_thresholds = candidate_thresholds(image)

    # segment image for each candidate threshold
    @dask.delayed(pure=True)
    def calculate_stats_for_threshold(grayscale, threshold):
        # change to THRESH_BINARY_INV?
        _, binarized = cv2.threshold(
            np.uint8(grayscale), threshold, 255, cv2.THRESH_BINARY)
        contours, _ = cv2.findContours(
            np.uint8(binarized), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)

        # find areas of objects for each segmented version
        areas = list(map(cv2.contourArea, contours))
        return [np.nanmedian(areas), binarized, contours, areas]
    out = {}
    # dask_thresholds must be computed because we need to iterate over it
    thresholds = dask_thresholds.compute()
    for threshold in thresholds:
        out[threshold] = calculate_stats_for_threshold(dask_gs, threshold)

    # return gs, {threshold: [median area, segmented img, contours, areas]}
    return dask_gs, out


def initial_threshold(image):
    dask_gs, dask_ima = median_areas(image)
    # needed to get max of delayed objects:
    img_median_areas = {key: value.compute()
                        for key, value in dask_ima.items()}

    @dask.delayed(nout=5, pure=True)
    def get_initial_threshold(median_areas):
        # get threshold with greatest median area
        img_initial_threshold = max(
            median_areas, key=lambda key: median_areas[key][0])

        median_area, img_segmented, contours, areas = median_areas[img_initial_threshold]
        return img_initial_threshold, median_area, img_segmented, contours, areas

    # return (gs, (initial threshold, median area, segmented img, contours, areas))
    return dask_gs, get_initial_threshold(img_median_areas)



import pymana.correction as correction
import numpy as np
import cv2
import dask
# from scipy import ndimage as ndi
# from skimage.segmentation import watershed
# from skimage.feature import peak_local_max


def separated(image):
    dask_gs, dask_threshold, dask_corrected_img = correction.corrected(
        image)

    @dask.delayed(pure=True)
    def expand_dimensions(grayscale):
        # the image is still grayscale but has 3 channels:
        return cv2.cvtColor(np.uint8(grayscale), cv2.COLOR_GRAY2BGR)
    img = expand_dimensions(dask_gs)

    @dask.delayed(nout=2, pure=True)
    def noise_removal(corrected):
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
        opening = cv2.morphologyEx(
            corrected, cv2.MORPH_OPEN, kernel, iterations=8)
        return kernel, opening
    kernel, opening = noise_removal(dask_corrected_img)

    @dask.delayed(pure=True)
    def get_background(opening, kernel):
        return cv2.dilate(opening, kernel, iterations=2)
    sure_bg = get_background(opening, kernel)

    @dask.delayed(nout=2, pure=True)
    def get_foreground(bg, threshold):
        dist_transform = cv2.distanceTransform(bg, cv2.DIST_L2, 5)
        _, fg_uncorrected = cv2.threshold(
            dist_transform, threshold, 255, cv2.THRESH_BINARY)
        contours, _ = cv2.findContours(
            np.uint8(fg_uncorrected), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        mean_area = np.mean(list(map(cv2.contourArea, contours)))
        sure_fg = correction.correct_mask(
            correction.classify_contours(contours, mean_area), fg_uncorrected)
        return dist_transform, np.uint8(sure_fg)
    dist_transform, sure_fg = get_foreground(sure_bg, dask_threshold)

    @dask.delayed(pure=True)
    def get_unknown(bg, fg):
        return cv2.subtract(bg, fg)

    unknown = get_unknown(sure_bg, sure_fg)

    @dask.delayed(nout=2, pure=True)
    def marker_watershed(bgr, fg, unknown):
        # distance = ndi.distance_transform_edt(bgr)
        # local_maxima = peak_local_max(distance,
        #                               indices=False)
        # markers = ndi.label(local_maxima)[0]
        # labels = watershed(-distance, markers)
        # return markers, labels

        _, markers = cv2.connectedComponents(fg)

        # Add one to all labels so that sure background is not 0, but 1
        markers += 1

        # Now, mark the region of unknown with zero
        markers[unknown == 255] = 0

        markers = cv2.watershed(bgr, markers)
        segmented = np.copy(bgr)
        segmented[markers == -1] = [255, 0, 0]

        return markers, segmented

    markers, segmented = marker_watershed(img, sure_fg, unknown)
    # segmented = marker_watershed(img, unknown)

    return sure_bg, sure_fg, dist_transform, unknown, segmented



#!/usr/bin/env python

import click
from click_aliases import ClickAliasedGroup
import dask
import napari
import numpy as np
import pprint
import pymana


@click.group(cls=ClickAliasedGroup)
@click.argument(u'image', type=click.Path(exists=True, dir_okay=False))
@click.pass_context
def cli(ctx, image):
    ctx.ensure_object(dict)
    ctx.obj['IMAGE'] = image


@cli.command()
@click.pass_context
def candidate_thresholds(ctx):
    # candidate-thresholds
    img = pymana.utils.svs2dask_array(ctx.obj['IMAGE'])[
                                      10000:12000, 10000:12000]
    dask_gs, dask_valid_thresholds = pymana.detection.candidate_thresholds(img)
    grayscale, valid_thresholds = dask.compute(dask_gs, dask_valid_thresholds)
    click.echo("Candidate thresholds: ")
    click.echo(pprint.pformat(valid_thresholds))
    with napari.gui_qt():
        _ = napari.view_image(np.uint8(grayscale),
                              name='grayscale',
                              rgb=False,
                              is_pyramid=False,
                              contrast_limits=[0, 255])
    return grayscale, valid_thresholds


@cli.command()
@click.pass_context
def median_areas(ctx):
    # median-areas
    img = pymana.utils.svs2dask_array(ctx.obj['IMAGE'])[
                                      10000:12000, 10000:12000]
    dask_gs, img_median_areas = pymana.detection.median_areas(img)
    median_areas_no_img = {key: value[0].compute()
                           for key, value in img_median_areas.items()}
    click.echo("Median object area for each candidate threshold: ")
    click.echo(pprint.pformat(median_areas_no_img))
    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(np.uint8(dask_gs.compute()),
                         name='grayscale',
                         rgb=False,
                         is_pyramid=False,
                         contrast_limits=[0, 255])
        for key, value in img_median_areas.items():
            viewer.add_image(np.uint8(value[1].compute()),
                             name=f'threshold = {key}',
                             rgb=False,
                             is_pyramid=False,
                             contrast_limits=[0, 255])

    return median_areas_no_img


@cli.command(aliases=['raw-nuclei-mask'])
@click.pass_context
def initial_threshold(ctx):
    # initial-threshold
    img = pymana.utils.svs2dask_array(ctx.obj['IMAGE'])[
                                      10000:12000, 10000:12000]
    dask_gs, (dask_it, dask_hma, dask_si, _, _) \
        = pymana.detection.initial_threshold(img)

    grayscale, img_initial_threshold, \
        highest_median_area, segmented_img = dask.compute(dask_gs, dask_it,
                                                          dask_hma, dask_si)
    click.echo("Initial threshold: ")
    click.echo(img_initial_threshold)
    click.echo("Median area for initial threshold: ")
    click.echo(highest_median_area)
    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(np.uint8(grayscale),
                         name='grayscale',
                         rgb=False,
                         is_pyramid=False,
                         contrast_limits=[0, 255])
        viewer.add_image(segmented_img,
                         name=f'threshold = {img_initial_threshold}',
                         rgb=False,
                         is_pyramid=False,
                         contrast_limits=[0, 255])
    return img_initial_threshold, highest_median_area, segmented_img


@cli.command()
@click.pass_context
def area_based_corrected(ctx):
    # area-based-corrected
    img = pymana.utils.svs2dask_array(ctx.obj['IMAGE'])[
                                      10000:12000, 10000:12000]
    dask_gs, dask_threshold, dask_is = pymana.correction.corrected(img)
    grayscale, threshold, img_segmented = dask.compute(
        dask_gs, dask_threshold, dask_is)
    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(np.uint8(grayscale),
                         name='grayscale',
                         rgb=False,
                         is_pyramid=False,
                         contrast_limits=[0, 255])
        viewer.add_image(img_segmented,
                         name=f'threshold = {threshold}',
                         rgb=False,
                         is_pyramid=False,
                         contrast_limits=[0, 255])


@cli.command()
@click.pass_context
def nuclei_separated(ctx):
    # nuclei-separated
    img = pymana.utils.svs2dask_array(ctx.obj['IMAGE'])[
                                      10000:12000, 10000:12000]
    dask_bg, dask_fg, dask_dt, dask_unknown, dask_corrected \
        = pymana.separation.separated(img)
    bg, fg, dist_transform, unknown, corrected_img = dask.compute(
        dask_bg,
        dask_fg,
        dask_dt,
        dask_unknown,
        dask_corrected)
    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(bg,
                         name='background',
                         rgb=False,
                         is_pyramid=False,
                         contrast_limits=[0, 255])
        viewer.add_image(fg,
                         name='foreground',
                         rgb=False,
                         is_pyramid=False,
                         contrast_limits=[0, 255])
        viewer.add_image(dist_transform,
                         name='dist_transform',
                         rgb=False,
                         is_pyramid=False,
                         contrast_limits=[0, 255])
        viewer.add_image(unknown,
                         name='unknown',
                         rgb=False,
                         is_pyramid=False,
                         contrast_limits=[0, 255])
        viewer.add_image(corrected_img,
                         name='final',
                         rgb=True,
                         is_pyramid=False,
                         contrast_limits=[0, 255])


if __name__ == '__main__':
    cli()



import pymana.detection as detection
import dask
import numpy as np
import cv2


def classify_contours(contours, mean):
    small, medium, large = [], [], []
    for contour in contours:
        if cv2.contourArea(contour) <= 0.25 * mean:
            small.append(contour)
        elif cv2.contourArea(contour) > 5 * mean:
            large.append(contour)
        else:
            medium.append(contour)
    return {'small': small, 'medium': medium, 'large': large}


def correct_mask(classifications, binarized):
    small_contours = classifications['small']
    small_contour_mask = np.ones_like(binarized, dtype=np.uint8)
    for small_contour in small_contours:
        cv2.drawContours(small_contour_mask, [small_contour], -1, 0, -1)

    corrected_mask = cv2.bitwise_and(
        binarized, binarized, mask=small_contour_mask)
    return cv2.bitwise_not(corrected_mask)


def corrected(image):
    dask_gs, (dask_it, _, _, dask_contours, dask_areas) \
        = detection.initial_threshold(image)
    # paper does not mention if mean should be recalculated for each iteration
    initial_contours, areas = dask.compute(dask_contours, dask_areas)
    mean_area = np.mean(areas)
    classifications = classify_contours(initial_contours, mean_area)

    @dask.delayed(pure=True, nout=2)
    def correct_threshold(grayscale, initial_classifications, initial_threshold, initial_mean_area):
        corrected_threshold = initial_threshold
        classifications = initial_classifications
        np_gs = np.uint8(grayscale)
        mean_area = initial_mean_area

        def iterate(corrected_threshold):
            _, binarized = cv2.threshold(
                np_gs,
                corrected_threshold,
                255,
                cv2.THRESH_BINARY_INV)
            contours, _ = cv2.findContours(
                binarized, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
            # mean_area = np.mean(list(map(cv2.contourArea, contours)))
            classifications = classify_contours(contours, mean_area)
            return binarized, contours, classifications
        binarized, contours, classifications = iterate(corrected_threshold)
        while len(classifications['large']) > 0 and corrected_threshold > 1:
            # use Bayesian optimization to do this more efficiently
            corrected_threshold -= 1  # -1 just for testing
            binarized, contours, classifications = iterate(corrected_threshold)
        return corrected_threshold, binarized

    corrected_threshold, binarized = correct_threshold(
        dask_gs,
        classifications,
        dask_it,
        mean_area
    )

    corrected_mask = dask.delayed(correct_mask, pure=True)(
        classifications, binarized)

    return dask_gs, corrected_threshold, corrected_mask



import openslide
import openslide.deepzoom as deepzoom
import numpy as np
import dask
import dask.array as da
import dask_image.imread


def svs2dask_array(svs_file, tile_size=1000, overlap=0, remove_last=True, allow_unknown_chunksizes=False):
    """Convert SVS, TIF or TIFF to dask array.
    Parameters
    ----------
    svs_file : str
            Image file.
    tile_size : int
            Size of chunk to be read in.
    overlap : int
            Do not modify, overlap between neighboring tiles.
    remove_last : bool
            Remove last tile because it has a custom size.
    allow_unknown_chunksizes : bool
            Allow different chunk sizes, more flexible, but slowdown.
    Returns
    -------
    arr : dask.array.Array
            A Dask Array representing the contents of the image file.
    >>> arr = svs2dask_array(svs_file, tile_size=1000, overlap=0, remove_last=True, allow_unknown_chunksizes=False)
    >>> arr2 = arr.compute()
    >>> arr3 = to_pil(cv2.resize(arr2, dsize=(1440, 700), interpolation=cv2.INTER_CUBIC))
    >>> arr3.save(test_image_name)
    """
    # https://github.com/jlevy44/PathFlowAI/blob/master/pathflowai/utils.py
    img = openslide.open_slide(svs_file)
    if type(img) is openslide.OpenSlide:
        gen = deepzoom.DeepZoomGenerator(
            img, tile_size=tile_size, overlap=overlap, limit_bounds=True)
        max_level = len(gen.level_dimensions) - 1
        n_tiles_x, n_tiles_y = gen.level_tiles[max_level]

        @dask.delayed(pure=True)
        def get_tile(level, column, row):
            tile = gen.get_tile(level, (column, row))
            return np.array(tile).transpose((1, 0, 2))

        sample_tile_shape = get_tile(max_level, 0, 0).shape.compute()
        rows = range(n_tiles_y - (0 if not remove_last else 1))
        cols = range(n_tiles_x - (0 if not remove_last else 1))
        arr = da.concatenate([da.concatenate([da.from_delayed(get_tile(max_level, col, row), sample_tile_shape, np.uint8) for row in rows],
                                             allow_unknown_chunksizes=allow_unknown_chunksizes, axis=1) for col in cols], allow_unknown_chunksizes=allow_unknown_chunksizes).transpose([1, 0, 2])
        return arr
    else:  # img is instance of openslide.ImageSlide
        return dask_image.imread.imread(svs_file)
        # return dask.array.from_array(np.asarray(img.get_thumbnail(img.dimensions)))



__version__ = '0.1.6'

from .detection import candidate_thresholds, median_areas, initial_threshold
from .correction import corrected
from .separation import separated
from .utils import svs2dask_array

__all__ = ['candidate_thresholds', 'median_areas', 'initial_threshold',
           'corrected', 'classify_contours', 'correct_mask',
           'separated', 'svs2dask_array']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pymana']

package_data = \
{'': ['*']}

install_requires = \
['click-aliases>=1.0.1,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'dask-image>=0.2.0,<0.3.0',
 'dask[array]>=2.16.0,<3.0.0',
 'napari>=0.3.1,<0.4.0',
 'numpy>=1.18.4,<2.0.0',
 'opencv-python-headless==4.2.0.34',
 'openslide-python>=1.1.1,<2.0.0',
 'scikit-image>=0.17.2,<0.18.0',
 'sympy>=1.5.1,<2.0.0']

entry_points = \
{'console_scripts': ['pymana = pymana.cli:cli']}

setup_kwargs = {
    'name': 'pymana',
    'version': '0.1.6',
    'description': '',
    'long_description': 'pymana\n==============\n\n:code:`pymana.separation.separated` does not work!\n',
    'author': 'Sumanth Ratna',
    'author_email': 'sumanthratna@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sumanthratna/pymana',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
